/*
Module : MfcCDDB.CPP
Purpose: Defines the implementation for a collection of MFC classes to wrap access to CDDB
Created: PJN / 24-02-1999
History: PJN / 19-05-1999 1. Added support for MOTD (Message of the Day)
                          2. Added support for submit
                          3. Fixed a number of places where calling the TRACE
                          function was causing an access violation
                          4. Fixed a bug in GetErrorMessage which was causing
                          SDK errors to return the string "The operation completed successfully"
         PJN / 14-07-1999 1. Fixed two potential access violations in the CCDDB::ReadResponse function.
         PJN / 29-11-1999 Fixed a number of bugs when compiled with VC 6.0
         PJN / 17-01-2000 Fixed a nmber of compiler errors which seem to have crept into v1.12
         PJN / 23-01-2000 1. The code has now been fully tested with the open source alternative to CDDB namely 
                          freedb.freedb.org. CDDB has become more commercialized of late, since its introduction 
                          of a binary SDK with closed validation and new licensening rules. I also believe that 
                          CDDB will stop access sometime in 2000 to its databases via non CDDB originated code.
                          Therefore I would suggest that developers using MfcCDDB should provide users with a 
                          configuration dialog to allow the CDDB server to be specified, that way you will not 
                          have irate customers ringing you when CDDB stops allowing access to it. You can use 
                          this open source server instead of the CDDB one in the MfcCDDB code by using 
                          "freedb.freedb.org" instead of "cddb.cddb.com" in the call to CCDDB::Sites and by 
                          specifying CDDB protocol version 3 in the function SetCDDBProtocolVersion.
         PJN / 13-02-2000 1. Now fully supports access when behind a HTTP proxy.
                          2. Also now includes support for proxy authentication
         PJN / 07-05-2000 Fixed a problem in GetUserName and GetHostName with embedded spaces
         PJN / 17-05-2000 Now handles CDDB records which have more than one DTITLE line
         PJN / 30-05-2000 Now prevents creation of empty extended track tags (EXTT=)
         PJN / 05-09-2000 Modified CCDDB::GetTrackPositions to fail when the CD does not contain any audio tracks
         PJN / 09-08-2001 Fixed a bug where the first track title was not being parsed correctly. Thanks to Jared 
                          Allen for spotting this.
         PJN / 06-10-2001 Major update to bring code into line with freedb standard. As mentioned in previous 
                          updates you can no longer use MFCCDDB with the commercial CDDB web site at Gracenote
                          (formerly CDDB). Instead now the code can only be used with the open source alternative
                          of freedb. Updates include:
                          1. Updated copyright information
                          2. Fixed calculation the extra disc length parameter, which must be specified in the query string.
                          3. Fixed bug where discid for submissions was having any leading 0's stripped.
                          4. Updated m_sProductVersion as it was out of date.
                          5. Changed default server in Sites method to "freedb.freedb.org"
                          6. Removed comment line which says "Copyright ... CDDB Inc" from Submit method
                          7. Updated links in documentation to be www.freedb.org
                          8. Fixed a bug in ReadResponse when an error occurs reading from the socket
                          9. Reworked the ReadResponse method to be more inline with my W3MFC socket class
                          10. Updated the sample app to display the last command response when a function fails
                          11. framework now works with version 5 of the CDDB protocol. This gives two extra
                          fields in the CDDB record class namely Year and Genre.
                          12. Complete rewrite of the ParseRecordBody to improve readibilty of the function. This 
                          means that the code now correctly supports all text record values which span multiple 
                          lines
                          13. Made m_ExtendedData member of CCDDBRecord a CString instead of a string array
                          14. Fixed a bug in CCDDB::ParseKeywordLine
                          15. Fixed bug when separating artist and title from DTITLE lines
                          16. Fixed an infinite recursion bug in ReplaceSubString
         PJN / 02-02-2002 1. Fixed a minor parsing bug in CCDDB::GetNextLine. Thanks to Thomas Lang for this fix.
         PJN / 24-06-2002 1. Now fully supports local access to database. The standard and alternate form are
                          both fully supported.
                          2. Optimized #include of 2 files in mfccddb.cpp
         PJN / 03-08-2003 1. Fixed a bug when the MCI device was not being closed if the inserted CD did not contain
                          an Audio CD. Thanks to Christian Inkster for reporting this problem.
         PJN / 03-11-2003 1. Simplified the code in CCDDB::ReadResponse. Thanks to Clarke Brunt for reporting this issue.
         PJN / 20-08-2004 1. General tidy up of the GetUserName and GetHostName functions.
         PJN / 29-06-2005 1. Updated copyright details.
                          2. Updated comments regarding how to include headers into your pre compiled header
                          3. Reviewed all TRACE statements for correctness
                          4. Updates to the GetUserName and GetHostName functions to exclude high ascii characters.
                          Thanks to Stefan Becker for providing the fix for this bug.
                          5. Removed unreferrenced variable warning in CCDDB::FormSubmitLine
                          6. Fixed a number of level 4 warnings when code is compiled in Visual Studio .NET 2003
                          7. Fixed various warnings when code is compiled with /Wp64 compiler setting in
                          Visual Studio .NET 2003. Please note that to achieve this, the code now requires a recentish
                          Platform SDK to be installed and configured if the code is compiled on Visual C++ 6
         PJN / 01-01-2007 1. Updated copyright details.
                          2. Addition of a MFCCDDB_EXT_CLASS preprocessor macro to allow the classes to be more easily 
                          added to an extension dll
                          3. Updated the code to clean compile in VC 2005
                          4. Updated documentation to use the same style as the web site
                          5. Default timeout is now 60 seconds for release builds and 90 seconds for debug builds
                          6. Reviewed all CString::Format calls for correctness
                          7. Optimized CCDDBSite constructor code
                          8. Optimized CCDDBQueryResult constructor code
                          9. Optimized CCDDBTrackPosition constructor code
                          10. Optimized CCDDBRecord constructor code
                          11. Optimized CCDDB constructor code
                          12. Removed unused CCDDB destructor
                          13. Fixed Unicode issue in CCDDB::GetCDROMDrives
                          14. Updated usage of CString::TrimLeft and CString::TrimRight in the code
                          15. Replaced usage of CCDDB::ReplaceSubString with CString::Replace
                          16. Hostname can now be customized via a SetHostName method
                          17. m_sLastCommandResponse is now not reset when a response is successfully parsed
                          18. UserName can now be customized via a SetUserName method
                          19. CCDDBSite::Parse now does not fail if protocol is not "http"
                          20. Reworked the classes to be exception based along the lines of the author's CPJNSMTPConnection
                          class. This should improve the robustness of client application code. Please note that because
                          of this redesign, you should reexamine all client code to make sure it catches the new
                          exceptions thrown from the classes as appropriate
                          21. Fixed a bug in CCDDB::SplitSemiColonLine where 1 character length values would not be 
                          parsed correctly
                          22. Added a constructor to the class CCDDBStatus
                          23. Fixed a bug in CCDDB::ParseCommentLine where 1 character length values would not be
                          parsed correctly
                          24. Fixed a bug in CCDDB::ReadResponse where it had problems parsing the body of the response
                          for query and submit requests if the body arrives in a second packet from the network
                          25. CCDDB::ParseRecordBodyFromFile now allows other processes to read the file which it is
                          currently attempting to parse
                          26. Optimized code in CCDDB::Read(const CString& sDirectory... which operates on the passed in
                          dwDiscID parameter
                          27. Optimized code in CCDDB::ParseRecordBodyFromFile to handle parsing the non-standard (Windows)
                          format of CDDB record format
         PJN / 12-01-2008 1. Updated copyright details.
                          2. Updated sample app code to remove VC 6 style classwizard comments
                          3. Updated the sample app to clean compile on VC 2008
                          4. CCDDBException::GetErrorMessage now uses the FORMAT_MESSAGE_IGNORE_INSERTS flag. For more 
                          information please see Raymond Chen's blog at 
                          http://blogs.msdn.com/oldnewthing/archive/2007/11/28/6564257.aspx. Thanks to Alexey Kuznetsov for 
                          reporting this issue.
                          5. CCDDBException::GetErrorMessage now uses Checked::tcsncpy_s if compiled using VC 2005 or 
                          later.
         PJN / 31-08-2008 1. Code now compiles cleanly using Code Analysis (/analyze)
                          2. Updated code to compile correctly using _ATL_CSTRING_EXPLICIT_CONSTRUCTORS define
                          3. The code now only supports VC 2005 or later. 
                          4. Added support for parsing CCDDBSite::m_sProtocol
                          5. Code now uses ATL's support for base64 encoding
                          6. The code now uses the author's CWSocket to encapsulate socket functionality
                          7. Fixed a bug in the calculation of the total track length in CCDDB::ComputeDiscID. Thanks
                          to Tom van Dijck for reporting this bug.
                          8. General tidy up of all the parsing code in the classes
                          9. The classes now supports v6 Protocol level (if the code is compiled for Unicode)
                          10. GetTrackPositions has been renamed to GetTrackPositionsUsingMCI.
                          11. Addition of a GetTrackPositionsUsingDeviceIoControl using the API DeviceIoControl and the 
                          IOCTL code of IOCTL_CDROM_READ_TOC to get the track positions. This method works on all 
                          versions of Windows based on the NT Kernel (i.e. NT, 2000, XP, 2003, Vista and 2008). It
                          should also work for multi-session CD's i.e. CD's which have data sessions as well as an
                          audio session. 
                          12. ComputeDiscID has been renamed to ComputeDiscIDUsingMCI.
                          13. Addition of a ComputeDiscIDUsingDeviceIoControl method.
                            
Copyright (c) 1999 - 2008 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////////////  Includes  //////////////////////////////////

#include "stdafx.h"
#ifndef _INC_MMSYSTEM
#pragma message("To avoid this message, please put mmsystem.h in your pre compiled header (normally stdafx.h)")
#include <mmsystem.h>
#endif
#ifndef NETCONS_INCLUDED
#pragma message("To avoid this message, please put lmcons.h in your pre compiled header (normally stdafx.h)")
#include <lmcons.h>
#endif
#ifndef __ATLENC_H__
#pragma message("To avoid this message, please put atlenc.h in your pre compiled header (normally stdafx.h)")
#include <atlenc.h>
#endif
#include "MfcCDDB.h"
#include "resource.h"


//////////////////////////////// Statics / Macros /////////////////////////////

//Automatically pull in the library winmm.lib since we
//need MCI to retrieve the cdaudio track details
#pragma comment(lib, "winmm.lib")

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

#ifndef METHOD_BUFFERED
#define METHOD_BUFFERED 0
#endif

#ifndef FILE_DEVICE_CD_ROM
#define FILE_DEVICE_CD_ROM 0x00000002
#endif

#ifndef IOCTL_CDROM_BASE
#define IOCTL_CDROM_BASE FILE_DEVICE_CD_ROM
#endif

#ifndef CTL_CODE
#define CTL_CODE( DeviceType, Function, Method, Access ) (                 \
    ((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method) \
)
#endif

#ifndef FILE_READ_ACCESS
#define FILE_READ_ACCESS 0x00000001
#endif

#ifndef IOCTL_CDROM_READ_TOC
#define IOCTL_CDROM_READ_TOC CTL_CODE(IOCTL_CDROM_BASE, 0x0000, METHOD_BUFFERED, FILE_READ_ACCESS)
#endif

#ifndef TRACK_DATA
typedef struct _TRACK_DATA 
{
  UCHAR Reserved;
  UCHAR Control : 4;
  UCHAR Adr : 4;
  UCHAR TrackNumber;
  UCHAR Reserved1;
  UCHAR Address[4];
} TRACK_DATA, *PTRACK_DATA;
#endif

#ifndef MAXIMUM_NUMBER_TRACKS
#define MAXIMUM_NUMBER_TRACKS 100
#endif

#ifndef CDROM_TOC
typedef struct _CDROM_TOC 
{
  UCHAR  Length[2];
  UCHAR  FirstTrack;
  UCHAR  LastTrack;
  TRACK_DATA  TrackData[MAXIMUM_NUMBER_TRACKS];
} CDROM_TOC, *PCDROM_TOC;
#endif


////////////////////////////////// Implementation /////////////////////////////

CCDDBException::CCDDBException(HRESULT hr, const CString& sLastResponse) : m_bMMError(FALSE),
                                                                           m_hr(hr), 
                                                                           m_sLastResponse(sLastResponse)
{
}

CCDDBException::CCDDBException(DWORD dwError, DWORD dwFacility, const CString& sLastResponse) : m_bMMError(FALSE),
                                                                                                m_hr(MAKE_HRESULT(SEVERITY_ERROR, dwFacility, dwError)),
                                                                                                m_sLastResponse(sLastResponse)
{
}

CCDDBException::CCDDBException(BOOL bMMError, DWORD dwError, const CString& sLastResponse) : m_bMMError(bMMError),
                                                                                             m_hr(MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, dwError)),
                                                                                             m_sLastResponse(sLastResponse)
{
}

BOOL CCDDBException::GetErrorMessage(LPTSTR pstrError, UINT nMaxError, PUINT pnHelpContext)
{
  //Validate our parameters
	ASSERT(pstrError != NULL && AfxIsValidString(pstrError, nMaxError));

	if (pnHelpContext != NULL)
		*pnHelpContext = 0;

  //What will be the return value from this function (assume the worst)
  BOOL bSuccess = FALSE;
  
  //If would be great if there was a FACILITY_MMSYSTEM, but there isn't, so we carry around an extra boolean
  //in the exception class which signifies if it is a MMSYSTEM error
  if (m_bMMError)
  {
    bSuccess = mciGetErrorString(HRESULT_CODE(m_hr), pstrError, nMaxError);
    if (bSuccess == FALSE)
      *pstrError = _T('\0');
  }
  else if (HRESULT_FACILITY(m_hr) == FACILITY_ITF)
  {
    bSuccess = TRUE;
  
    //Simply load up the string from the string table
    CString sError;
    AfxFormatString1(sError, HRESULT_CODE(m_hr), m_sLastResponse);
    Checked::tcsncpy_s(pstrError, nMaxError, sError, _TRUNCATE);  
  }
  else
  {
  	LPTSTR lpBuffer;
	  DWORD dwReturn = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
			                             NULL, HRESULT_CODE(m_hr), MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT),
			                             reinterpret_cast<LPTSTR>(&lpBuffer), 0, NULL);

	  if (dwReturn == 0)
		  *pstrError = _T('\0');
	  else
	  {
	    bSuccess = TRUE;
      Checked::tcsncpy_s(pstrError, nMaxError, lpBuffer, _TRUNCATE);  
		  LocalFree(lpBuffer);
	  }
  }

	return bSuccess;
}

CString CCDDBException::GetErrorMessage()
{
  CString rVal;
  LPTSTR pstrError = rVal.GetBuffer(4096);
  GetErrorMessage(pstrError, 4096, NULL);
  rVal.ReleaseBuffer();
  return rVal;
}

IMPLEMENT_DYNAMIC(CCDDBException, CException)

#ifdef _DEBUG
void CCDDBException::Dump(CDumpContext& dc) const
{
  //Let the base class do its thing
	CObject::Dump(dc);

	dc << _T("m_hr = ") << m_hr << _T("\n");
}
#endif


CCDDBSite::CCDDBSite(BOOL bRetrieve) : m_nPort(80),
                                       m_bNorthing(TRUE),
                                       m_nLatitudeMinutes(0),
                                       m_bEasting(TRUE),
                                       m_nLongitudeMinutes(0)
{
  if (bRetrieve)
    m_sAddress = _T("/~cddb/cddb.cgi");
  else
    m_sAddress = _T("/~cddb/submit.cgi");
}

CCDDBSite::CCDDBSite(const CCDDBSite& site)
{
  *this = site;
}

CCDDBSite& CCDDBSite::operator=(const CCDDBSite& site)
{
  m_sSite             = site.m_sSite;
  m_nPort             = site.m_nPort;
  m_sAddress          = site.m_sAddress;
  m_bNorthing         = site.m_bNorthing;
  m_nLatitudeMinutes  = site.m_nLatitudeMinutes;
  m_bEasting          = site.m_bEasting;
  m_nLongitudeMinutes = site.m_nLongitudeMinutes;
  m_sDescription      = site.m_sDescription;
  m_sProtocol         = site.m_sProtocol;
  
  return *this;
}



CCDDBQueryResult::CCDDBQueryResult() : m_dwDiscID(0)
{
}

CCDDBQueryResult::CCDDBQueryResult(const CCDDBQueryResult& result)
{
  *this = result;
}

CCDDBQueryResult& CCDDBQueryResult::operator=(const CCDDBQueryResult& result)
{
  m_sCategory = result.m_sCategory;
  m_dwDiscID  = result.m_dwDiscID;
  m_sArtist   = result.m_sArtist;
  m_sTitle    = result.m_sTitle;

  return *this;
}


CCDDBTrackPosition::CCDDBTrackPosition() : m_nMinute(0),
                                           m_nSecond(0),
                                           m_nFrame(0)

{
}

CCDDBTrackPosition::CCDDBTrackPosition(const CCDDBTrackPosition& position)
{
  *this = position;
}

CCDDBTrackPosition& CCDDBTrackPosition::operator=(const CCDDBTrackPosition& position)
{
  m_nMinute = position.m_nMinute;
  m_nSecond = position.m_nSecond;
  m_nFrame  = position.m_nFrame;

  return *this;
}


CCDDBStatus::CCDDBStatus() : m_nCurrentProtocol(0),
                             m_nMaxProtocol(0),
                             m_bGetsAllowed(FALSE), 
                             m_bUpdatesAllowed(FALSE),
                             m_bPostingAllowed(FALSE),
                             m_bQuotes(FALSE),
                             m_nCurrentUsers(0),  
                             m_nMaxUsers(0),
                             m_bStripExtended(FALSE),
                             m_nDatabaseEntries(0)
{
}


CCDDBRecord::CCDDBRecord() : m_nDiskLength(0),
                             m_nDatabaseRevision(0),
                             m_dwDiscID(0),
                             m_nYear(0)
{
}

CCDDBRecord::CCDDBRecord(const CCDDBRecord& record)
{
  *this = record;
}

CCDDBRecord& CCDDBRecord::operator=(const CCDDBRecord& result)
{
  m_TrackOffsets.Copy(result.m_TrackOffsets);
  m_nDiskLength       = result.m_nDiskLength;
  m_nDatabaseRevision = result.m_nDatabaseRevision;
  m_sClientName       = result.m_sClientName;
  m_sClientVersion    = result.m_sClientVersion;
  m_sClientComments   = result.m_sClientComments;
  m_dwDiscID          = result.m_dwDiscID;
  m_sArtist           = result.m_sArtist;
  m_sTitle            = result.m_sTitle;
  m_TrackTitles.Copy(result.m_TrackTitles);
  m_ExtendedData      = result.m_ExtendedData;
  m_ExtendedTrackData.Copy(result.m_ExtendedTrackData);
  m_PlayOrder.Copy(result.m_PlayOrder);
  m_nYear             = result.m_nYear;
  m_sGenre            = result.m_sGenre;

  return *this;
}


CCDDB::CCDDB() : m_sProductName(_T("MfcCDDB")), 
                 m_sProductVersion(_T("1.31")),
               #ifdef _UNICODE
                 m_nProtocolVersion(6),
               #else  
                 m_nProtocolVersion(5),
               #endif
                 m_nProxyPort(80)

{
#ifdef _DEBUG
  m_dwTimeout = 90000; //default timeout is set to 90 seconds for debug
#else
  m_dwTimeout = 60000;  //default timeout of 60 seconds for normal release code
#endif
  SetHostname(_T("auto"));
  SetUserName(_T("auto"));
}

void CCDDB::SetCDDBProtocolVersion(int nProtocolVersion) 
{
  //Validate our parameters
  ASSERT(nProtocolVersion >= 3 && nProtocolVersion <= 6); //MfcCDDB source code only supports v3, v4, v5 & v6.
  
  m_nProtocolVersion = nProtocolVersion; 
}

DWORD CCDDB::ComputeDiscID(const CArray<CCDDBTrackPosition, CCDDBTrackPosition&>& tracks)
{
  int nTracks = static_cast<int>(tracks.GetSize() - 1); //Number of tracks is 1 less than the size
                                                        //of the array as it also contains the lead 
                                                        //out position
  //Validate our parameters
  ASSERT(nTracks > 0);

  //Iterate across all the tracks
  int n = 0;
  for (int i=0; i<nTracks; i++)
  {
    int sum = 0;
    int j = tracks[i].m_nMinute*60 + tracks[i].m_nSecond;
    while (j > 0)
    {
      sum += j%10;
      j /=10;
    }
    n += sum;
  }

  //Compute total track length in seconds
  int nTotalTrackLength = tracks[nTracks].m_nMinute*60 + tracks[nTracks].m_nSecond - tracks[0].m_nMinute*60 - tracks[0].m_nSecond;

  //Compute DISC ID
  DWORD dwDiscID = ((n % 0xFF) << 24 | nTotalTrackLength << 8 | nTracks);
  return dwDiscID;
}

void CCDDB::GetTrackPositionsUsingDeviceIoControl(CArray<CCDDBTrackPosition, CCDDBTrackPosition&>& tracks, LPCTSTR pszDrive)
{
  //Remove any tracks already in the array
  tracks.RemoveAll();

  //Convert the drive format layout of the "pszDrive" parameter to the device format
  CString sDevice(_T("\\\\.\\\\"));
  sDevice += pszDrive;
  
  //first open the device
  ATL::CHandle handle(CreateFile(sDevice, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, 0));
  if (handle == INVALID_HANDLE_VALUE)
    ThrowCDDBException(GetLastError(), FACILITY_WIN32);

  //Issue the IOCTL to read the toc
  CDROM_TOC toc;
  memset(&toc, 0, sizeof(toc));
  DWORD dwBytesReturn = 0;
  if (!DeviceIoControl(handle, IOCTL_CDROM_READ_TOC, NULL, 0, &toc, sizeof(toc), &dwBytesReturn, NULL))
    ThrowCDDBException(GetLastError(), FACILITY_WIN32);
  
  //If successful, fill in the output parameter array
  int nTotalTracks = toc.LastTrack - toc.FirstTrack + 1;
  tracks.SetSize(nTotalTracks + 1);  
  for (UCHAR nTrack=toc.FirstTrack; nTrack<=(toc.LastTrack+1); nTrack++) //Note this code treats filling in the lead out position just like any other track
  {
    //Save the track position in MSF format
    CCDDBTrackPosition trackPosition;
    trackPosition.m_nMinute = toc.TrackData[nTrack-toc.FirstTrack].Address[0]*60 + toc.TrackData[nTrack-toc.FirstTrack].Address[1];
    trackPosition.m_nSecond = toc.TrackData[nTrack-toc.FirstTrack].Address[2];
    trackPosition.m_nFrame  = toc.TrackData[nTrack-toc.FirstTrack].Address[3];

    //Store the value in the array
    tracks.SetAt(nTrack-toc.FirstTrack, trackPosition);
  }
}

void CCDDB::GetTrackPositionsUsingMCI(CArray<CCDDBTrackPosition, CCDDBTrackPosition&>& tracks, LPCTSTR pszDrive)
{
  //Remove any tracks already in the array
  tracks.RemoveAll();

  //Open the specified "cdaudio" MCI device
  MCI_OPEN_PARMS mciOpenParms;
  mciOpenParms.lpstrDeviceType = _T("cdaudio");
  mciOpenParms.lpstrElementName = pszDrive;
  DWORD dwError = mciSendCommand(0, MCI_OPEN, MCI_OPEN_SHAREABLE | MCI_OPEN_TYPE | (pszDrive ? MCI_OPEN_ELEMENT : 0), reinterpret_cast<DWORD_PTR>(&mciOpenParms));
  if (dwError)
    ThrowCDDBMMException(dwError);

  //Set the time format to Minute/Second/Frame (MSF) format
  MCI_SET_PARMS mciSetParms;
  mciSetParms.dwTimeFormat = MCI_FORMAT_MSF;
  dwError = mciSendCommand(mciOpenParms.wDeviceID, MCI_SET, MCI_SET_TIME_FORMAT, reinterpret_cast<DWORD_PTR>(&mciSetParms));
  if (dwError)
  {
    //Dont forget to close the MCI device before we throw the exception
    mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);

    ThrowCDDBMMException(dwError);
  }

  //Get the total track count
  MCI_STATUS_PARMS mciStatusParms;
  mciStatusParms.dwItem = MCI_STATUS_NUMBER_OF_TRACKS;
  dwError = mciSendCommand(mciOpenParms.wDeviceID, MCI_STATUS, MCI_STATUS_ITEM, reinterpret_cast<DWORD_PTR>(&mciStatusParms));
  if (dwError)
  {
    //Dont forget to close the MCI device before we throw the exception
    mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);

    ThrowCDDBMMException(dwError);
  }
  int nTotalTracks = static_cast<int>(mciStatusParms.dwReturn);

  //check that disk contain audio track
  BOOL bAudioTrackPresent = FALSE;	
  for (int i=1; i<=nTotalTracks && !bAudioTrackPresent; i++)
  {                      
		mciStatusParms.dwItem = MCI_CDA_STATUS_TYPE_TRACK;
		mciStatusParms.dwTrack = i;
		dwError = mciSendCommand(mciOpenParms.wDeviceID, MCI_STATUS, MCI_STATUS_ITEM | MCI_TRACK, reinterpret_cast<DWORD_PTR>(&mciStatusParms));
		if (dwError)
		{
			//Dont forget to close the MCI device before we throw the exception
			mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);
			
			ThrowCDDBMMException(dwError);
		}
		if (mciStatusParms.dwReturn == MCI_CDA_TRACK_AUDIO)
			bAudioTrackPresent = TRUE;
	}
  if (!bAudioTrackPresent)  
  {
    //Dont forget to close the MCI device before we throw the exception
    mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);

    ThrowCDDBException(IDS_MFCCDDB_NO_AUDIO_TRACKS_PRESENT, FACILITY_ITF);
  } 

  //Iterate through all the tracks getting their starting position
  tracks.SetSize(nTotalTracks + 1);
  for (int i=1; i<=nTotalTracks; i++)
  {                      
    mciStatusParms.dwItem = MCI_STATUS_POSITION;
    mciStatusParms.dwTrack = i;
    dwError = mciSendCommand(mciOpenParms.wDeviceID, MCI_STATUS, MCI_STATUS_ITEM | MCI_TRACK, reinterpret_cast<DWORD_PTR>(&mciStatusParms));
    if (dwError)
    {
      //Dont forget to close the MCI device before we throw the exception
      mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);

      ThrowCDDBMMException(dwError);
    }

    //Save the track position in MSF format
    CCDDBTrackPosition trackPosition;
    trackPosition.m_nMinute = MCI_MSF_MINUTE(mciStatusParms.dwReturn);
    trackPosition.m_nSecond = MCI_MSF_SECOND(mciStatusParms.dwReturn);
    trackPosition.m_nFrame  = MCI_MSF_FRAME(mciStatusParms.dwReturn);

    //Store the value in the array
    tracks.SetAt(i-1, trackPosition);
  }

  //Get the last track's length
  mciStatusParms.dwItem = MCI_STATUS_LENGTH;
  mciStatusParms.dwTrack = nTotalTracks;
  dwError = mciSendCommand(mciOpenParms.wDeviceID, MCI_STATUS, MCI_STATUS_ITEM | MCI_TRACK, reinterpret_cast<DWORD_PTR>(&mciStatusParms));
  if (dwError)
  {
    //Dont forget to close the MCI device before we throw the exception
    mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);

    ThrowCDDBMMException(dwError);
  }
  
  //Compute lead-out track position
  DWORD dwLenM = MCI_MSF_MINUTE(mciStatusParms.dwReturn);
  DWORD dwLenS = MCI_MSF_SECOND(mciStatusParms.dwReturn);
  DWORD dwLenF = MCI_MSF_FRAME(mciStatusParms.dwReturn) + 1; //Fix MCI Windows bug according to CDDB Howto doc
  DWORD dwPosM = tracks[nTotalTracks-1].m_nMinute;
  DWORD dwPosS = tracks[nTotalTracks-1].m_nSecond;
  DWORD dwPosF = tracks[nTotalTracks-1].m_nFrame;

  //Compute lead out track position (in frame format)
  DWORD dwPos = (dwPosM*60*75) + (dwPosS*75) + dwPosF + (dwLenM*60*75) + (dwLenS*75) + dwLenF;

  //Convert dwPos back to MSF format
  CCDDBTrackPosition trackPosition;
  trackPosition.m_nFrame = dwPos % 75;
  dwPos /= 75;
  trackPosition.m_nSecond = dwPos % 60;
  dwPos /= 60;
  trackPosition.m_nMinute = dwPos;

  //And store in the array
  tracks.SetAt(nTotalTracks, trackPosition);

  //Dont forget to close the MCI device before we return
  mciSendCommand(mciOpenParms.wDeviceID, MCI_CLOSE, 0, 0);
}

DWORD CCDDB::ComputeDiscIDUsingMCI(LPCTSTR pszDrive)
{
  //Get the track details
  CArray<CCDDBTrackPosition, CCDDBTrackPosition&> tracks;
  GetTrackPositionsUsingMCI(tracks, pszDrive);
  
  //Compute the DISC ID now that we have got all the track information
  return ComputeDiscID(tracks);
}

DWORD CCDDB::ComputeDiscIDUsingDeviceIoControl(LPCTSTR pszDrive)
{
  //Get the track details
  CArray<CCDDBTrackPosition, CCDDBTrackPosition&> tracks;
  GetTrackPositionsUsingDeviceIoControl(tracks, pszDrive);
  
  //Compute the DISC ID now that we have got all the track information
  return ComputeDiscID(tracks);
}
    
void CCDDB::GetCDROMDrives(CStringArray& drives)
{
  //empty out the array
  drives.RemoveAll();

  //Iterate across all the drive letters to find out which ones are CDROMs
  for (int i=1; i<=26; i++)
  {
    CString sDrive;
    sDrive.Format(_T("%c:"), _T('A') + i - 1);
    if (GetDriveType(sDrive) == DRIVE_CDROM)
      drives.Add(sDrive);
  }
}

BOOL CCDDB::ReadResponse(CWSocket& socket, LPSTR pszTerminator, CStringA& sResponse, DWORD dwHint)
{
  //Retrieve the reponse using until we get the terminator or a timeout occurs
	BOOL bFoundTerminator = FALSE;
  BOOL bContinue = TRUE;
	while (!bFoundTerminator && bContinue)
	{
    //check the socket for readability
    try
    {
      if (!socket.IsReadible(m_dwTimeout))
      {
        //Hive away the last command reponse
        m_sLastCommandResponse = CA2T(sResponse); 
			  return FALSE;
      }
    }
    catch(CWSocketException* pEx)
    {
      DWORD dwError = pEx->m_nError;
      pEx->Delete();
			ThrowCDDBException(dwError, FACILITY_WIN32);
    }

		//receive the data from the socket
		char sBuf[4096];
		int nData = 0;
		try
		{
      nData = socket.Receive(sBuf, sizeof(sBuf) - 1);
    }
    catch(CWSocketException* pEx)
    {
      DWORD dwError = pEx->m_nError;
      pEx->Delete();
      ThrowCDDBException(dwError, FACILITY_WIN32);
    }  

    //Handle a graceful disconnect
    if (nData == 0)
      ThrowCDDBException(IDS_MFCCDDB_GRACEFUL_DISCONNECT, FACILITY_ITF, CString(CA2T(sResponse)));

    //NULL terminate the data received
    sBuf[nData] = '\0';
    
    //grow the response data
    sResponse += sBuf;      

    //Special case code for reading a Query and submit responses
    if (dwHint == ReadResponseQuery || dwHint == ReadResponseSubmit)
    {
      //Extract the HTTP body from the response;
      LPCSTR pszBody = FindHTTPBody(sResponse);
      if (pszBody && strlen(pszBody))
      {
        //From the HTTP body get the CDDB response code
        int nResponseCode = GetCDDBReponseCode(pszBody);

        //Only continue to receive data if the response code indicates more data is to be received
        if ((nResponseCode < 210 || nResponseCode > 219))
          bContinue = FALSE;          
      }
    }

    //Check to see if the terminator character(s) have been found
		bFoundTerminator = (strstr(sBuf, pszTerminator) != NULL);
	}

  //Hive away the last command response  
  m_sLastCommandResponse = sResponse;
    
  return TRUE;
}

void CCDDB::StripNonAlphaOrDigits(CString& sValue)
{
  int nLength = sValue.GetLength();
  for (int i=0; i<nLength; i++) 
  {
    TCHAR c = sValue.GetAt(i);
    BOOL bIsAlpha = (((c>=_T('A')) && (c<=_T('Z'))) || //Definitively between A-Z or a-z and nothing else!
                     ((c>=_T('a')) && (c<=_T('z')))); 
    BOOL bIsDigit = ((c>=_T('0')) && (c<=_T('9')));
    if (!(bIsAlpha || bIsDigit))
      sValue.SetAt(i, _T('_'));
  }
}

void CCDDB::SetUserName(const CString& sUserName)
{
	if (sUserName == _T("auto"))
	{
	  //Get the current windows user name
    TCHAR pszUser[UNLEN + 1];
    pszUser[0] = _T('\0');
    DWORD dwSize = UNLEN + 1;
    if (::GetUserName(pszUser, &dwSize))
      m_sUserName = pszUser;
    else
      m_sUserName = _T("Unknown");
  }
  else
    m_sUserName = sUserName;

  //convert string to alpha chars and digits only
  StripNonAlphaOrDigits(m_sUserName);
}

void CCDDB::SetHostname(const CString& sHostname)
{
	//retrieve the localhost name if we are supplied the magic "auto" host name
	if (sHostname == _T("auto"))
	{
		//get the current windows host name
		char szHostName[256];
		szHostName[0] = _T('\0');
		gethostname(szHostName, sizeof(szHostName));
		m_sHostName = CA2T(szHostName);
	}
  else
    m_sHostName = sHostname;
    
  //convert string to alpha chars and digits only
  StripNonAlphaOrDigits(m_sHostName);
}

CString CCDDB::GetHelloCommand()
{
  CString sCommand;
  sCommand.Format(_T("hello=%s+%s+%s+%s&proto=%d"), GetUserName().operator LPCTSTR(), GetHostName().operator LPCTSTR(), 
                  m_sProductName.operator LPCTSTR(), m_sProductVersion.operator LPCTSTR(), m_nProtocolVersion);
  return sCommand;
}

LPCSTR CCDDB::FindHTTPBody(LPCSTR pszResponse)
{
  //What will be the return value from this function
  LPCSTR pszData = NULL;

  if (pszResponse)
  {
    //Find the HTTP body
    pszData = strstr(pszResponse, "\r\n\r\n");
  
    //If found, skip over the 2 lines
    if (pszData)
      pszData += 4;
  }

  return pszData;  
}

int CCDDB::GetCDDBReponseCode(LPCSTR pszBody)
{
  //What will be the return value from this function
  int nResponse = -1;
  
  if (pszBody)
  {
    size_t nLength = strlen(pszBody);
    if (nLength >= 3)
    {
      char sNum[4];
      strncpy_s(sNum, sizeof(sNum), pszBody, 3);
      sNum[3] = '\0';
      nResponse = atoi(sNum);
    }
  }
  
  return nResponse;
}

BOOL CCDDB::SplitSemiColonLine(const CStringA& sLine, CStringA& sField, CStringA& sValue)
{
  //What will be the return value from this function (assume the worst)
  BOOL bSuccess = FALSE;

  //Do the actual parsing
  int nSeparator = sLine.Find(":");
  if (nSeparator != -1)
  {
    sField = sLine.Left(nSeparator);
    sValue = sLine.Right(sLine.GetLength() - nSeparator - 1);
  
    //Remove unwanted spaces 
    sField.Trim();
    sValue.Trim();

    bSuccess = TRUE;
  }

  return bSuccess;
}

void CCDDB::Sites(CArray<CCDDBSite, CCDDBSite&>& sites, const CCDDBSite& server)
{
  //Remove any entries which are in the sites array
  sites.RemoveAll();

  CWSocket socket;
  try
  {
    //Create the socket
    socket.Create();

    //Connect to the host on the specified port
    BOOL bUseProxy = !m_sProxyHost.IsEmpty();
    socket.Connect(bUseProxy ? m_sProxyHost : server.m_sSite, bUseProxy ? m_nProxyPort : server.m_nPort);

    //Form the specialized HTTP request to retrieve the sites data
    CString sRequest;
    if (!bUseProxy)
      sRequest.Format(_T("GET %s?cmd=sites&%s HTTP/1.0\r\n\r\n"), server.m_sAddress.operator LPCTSTR(), GetHelloCommand().operator LPCTSTR());
    else
      sRequest.Format(_T("GET http://%s%s?cmd=sites&%s HTTP/1.0\r\n%s\r\n\r\n"), server.m_sSite.operator LPCTSTR(), server.m_sAddress.operator LPCTSTR(), 
                      GetHelloCommand().operator LPCTSTR(), GetProxyHeader(server.m_sSite).operator LPCTSTR());

    //Send the request through the socket
    CT2A szRequest(sRequest);
    int nCmdLength = static_cast<int>(strlen(szRequest));
    socket.Send(szRequest, nCmdLength);
    
    //Read the response back through the socket
    CStringA sResponse;
    sResponse.Preallocate(4096);
    if (ReadResponse(socket, "\r\n.\r\n", sResponse))
    {
      //Extract the HTTP body from the response;
      LPCSTR pszBody = FindHTTPBody(sResponse);

      //Parse the sites info from the response
      if (!ParseSitesBody(pszBody, sites))
        ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_SITES_RESPONSE, m_sLastCommandResponse);
    }
    else
      ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_SITES_RESPONSE, m_sLastCommandResponse);
  }
  catch(CWSocketException* pEx)
  {
    DWORD dwError = pEx->m_nError;
    pEx->Delete();
    
    ThrowCDDBException(dwError, FACILITY_WIN32);    
  }
}

BOOL CCDDB::ParseSitesBody(LPCSTR pszBody, CArray<CCDDBSite, CCDDBSite&>& sites)
{
  //What will be the return value from this function (assume the worst)
  BOOL bSuccess = FALSE;

  //From the HTTP body get the CDDB response code
  int nResponseCode = GetCDDBReponseCode(pszBody);

  //If the response is in the expected range (21x) then parse the site data line by line
  if (nResponseCode >= 210 && nResponseCode <= 219)
  {
    //Skip over the current line
    LPCSTR pszLine = pszBody;
    CStringA sLine;
    GetNextLine(pszLine, sLine);
    
    //Parse each line
    while (pszLine)
    {
      //parse the current line and get the next line
      GetNextLine(pszLine, sLine);

      //If we can parse the current line then add it 
      //to the sites array
      if (sLine != ".")
      {
        CCDDBSite site;
        if (ParseSiteBodyHelper(site, sLine))
          sites.Add(site);
      }
    }
    
    bSuccess = TRUE;
  }
  
  return bSuccess;
}

void CCDDB::Sites(CArray<CCDDBSite, CCDDBSite&>& sites, const CString& sServer, const CString& sAddress, int nPort)
{
  //Form the site which we will be querying
  CCDDBSite server;
  server.m_sSite = sServer;
  server.m_nPort = nPort;
  server.m_sAddress = sAddress;
  
  //Let the other version do the heavy lifting
  Sites(sites, server);
}

void CCDDB::Categories(const CCDDBSite& server, CStringArray& categories)
{
  //Remove any entries which are in the categories array
  categories.RemoveAll();

  CWSocket socket;
  try
  {
    //Create the socket
    socket.Create();

    //Connect to the host on the specified port
    BOOL bUseProxy = !m_sProxyHost.IsEmpty();
    socket.Connect(bUseProxy ? m_sProxyHost : server.m_sSite, bUseProxy ? m_nProxyPort : server.m_nPort);
    
    //Form the specialized HTTP request to retreive the sites data
    CString sRequest;
    if (!bUseProxy)
      sRequest.Format(_T("GET %s?cmd=cddb+lscat&%s HTTP/1.0\r\n\r\n"), server.m_sAddress.operator LPCTSTR(), GetHelloCommand().operator LPCTSTR());
    else
      sRequest.Format(_T("GET http://%s%s?cmd=cddb+lscat&%s HTTP/1.0\r\n%s\r\n\r\n"), server.m_sSite.operator LPCTSTR(), server.m_sAddress.operator LPCTSTR(), 
                      GetHelloCommand().operator LPCTSTR(), GetProxyHeader(server.m_sSite).operator LPCTSTR());

    //Send the request through the socket
    CT2A szRequest(sRequest);
    int nCmdLength = static_cast<int>(strlen(szRequest));
    socket.Send(szRequest, nCmdLength);
    
    //Read the response back through the socket
    CStringA sResponse;
    sResponse.Preallocate(4096);
    if (ReadResponse(socket, "\r\n.\r\n", sResponse))
    {
      //Extract the HTTP body from the response;
      LPCSTR pszBody = FindHTTPBody(sResponse);

      //Parse the categories info from the response
      if (!ParseCategoriesBody(pszBody, categories))
        ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_CATEGORIES_RESPONSE, m_sLastCommandResponse);
    }
    else
      ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_CATEGORIES_RESPONSE, m_sLastCommandResponse);
  }
  catch(CWSocketException* pEx)
  {
    DWORD dwError = pEx->m_nError;
    pEx->Delete();
    
    ThrowCDDBException(dwError, FACILITY_WIN32);
  }
}

BOOL CCDDB::ParseCategoriesBody(LPCSTR pszBody, CStringArray& categories)
{
  //What will be the return value from this function (assume the worst)
  BOOL bSuccess = FALSE;

  //From the HTTP body get the CDDB response code
  int nResponseCode = GetCDDBReponseCode(pszBody);

  //If the response is in the expected range (21x) then parse the categories data line by line
  if (nResponseCode >= 210 && nResponseCode <= 219)
  {
    //Skip over the current line
    LPCSTR pszLine = pszBody;
    CStringA sLine;
    GetNextLine(pszLine, sLine);

    //Parse each line  
    while (pszLine)
    {
      //parse the current line and get the next line
      GetNextLine(pszLine, sLine);

      //Add the current line to the categories array
      if (sLine.GetLength() && sLine != ".")
        categories.Add(MapFromProtocolString(sLine));
    }

    bSuccess = TRUE;
  }

  return bSuccess;
}

void CCDDB::Status(const CCDDBSite& server, CCDDBStatus& status)
{
  //Empty out all the arrays in case they already have something in them
  status.m_Categories.RemoveAll();
  status.m_CategoryEntries.RemoveAll();
  status.m_PendingSites.RemoveAll();
  status.m_PendingEntries.RemoveAll();
 
  CWSocket socket;
  try
  {
    //Create the socket
    socket.Create();

    //Connect to the host on the specified port
    BOOL bUseProxy = !m_sProxyHost.IsEmpty();
    socket.Connect(bUseProxy ? m_sProxyHost : server.m_sSite, bUseProxy ? m_nProxyPort : server.m_nPort);

    //Form the specialized HTTP request to retreive the status data
    CString sRequest;
    if (!bUseProxy)
      sRequest.Format(_T("GET %s?cmd=stat&%s HTTP/1.0\r\n\r\n"), server.m_sAddress.operator LPCTSTR(), GetHelloCommand().operator LPCTSTR());
    else
      sRequest.Format(_T("GET http://%s%s?cmd=stat&%s HTTP/1.0\r\n%s\r\n\r\n"), server.m_sSite.operator LPCTSTR(), server.m_sAddress.operator LPCTSTR(), 
                      GetHelloCommand().operator LPCTSTR(), GetProxyHeader(server.m_sSite).operator LPCTSTR());

    //Send the request through the socket
    CT2A szRequest(sRequest);
    int nCmdLength = static_cast<int>(strlen(szRequest));
    socket.Send(szRequest, nCmdLength);

    //Read the response back through the socket
    CStringA sResponse;
    sResponse.Preallocate(4096);
    if (ReadResponse(socket, "\r\n.\r\n", sResponse))
    {
      //Extract the HTTP body from the response;
      LPCSTR pszBody = FindHTTPBody(sResponse);

      //Parse out the status from the body
      if (!ParseStatusBody(pszBody, status))
        ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_STATUS_RESPONSE, m_sLastCommandResponse);
    }
    else
      ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_STATUS_RESPONSE, m_sLastCommandResponse);
  }
  catch(CWSocketException* pEx)
  {
    DWORD dwError = pEx->m_nError;
    pEx->Delete();
    
    ThrowCDDBException(dwError, FACILITY_WIN32);
  }
}

BOOL CCDDB::ParseStatusBody(LPCSTR pszBody, CCDDBStatus& status)
{
  //What will be the return value from this function (assume the worst)
  BOOL bSuccess = FALSE;

  //From the HTTP body get the CDDB response code
  int nResponseCode = GetCDDBReponseCode(pszBody);

  //If the response is in the expected range (21x) then parse the status info line by line
  if (nResponseCode >= 210 && nResponseCode <= 219)
  {
    BOOL bCurrentlyParsingCategories = FALSE;
    BOOL bCurrentlyParsingSites = FALSE;
  
    //Skip over the current line
    LPCSTR pszLine = pszBody;
    CStringA sLine;
    GetNextLine(pszLine, sLine);
  
    //iterate across all lines
    while (pszLine)
    {
      //parse the current line and get the next line
      GetNextLine(pszLine, sLine);

      //Parse each line
      CStringA sField;
      CStringA sValue;
      if (bCurrentlyParsingCategories)
      {
        //Parse the category info
        if (sLine.Left(1) == ' ')
        {
          if (SplitSemiColonLine(sLine, sField, sValue))
          {
            status.m_Categories.Add(MapFromProtocolString(sField));
            int nValue = atoi(sValue);
            status.m_CategoryEntries.Add(nValue);
          }
        }
        else
          bCurrentlyParsingCategories = FALSE;
      }

      if (bCurrentlyParsingSites)
      {
        //Parse the pending sites info
        if (sLine.Left(1) == ' ')
        {
          if (SplitSemiColonLine(sLine, sField, sValue))
          {
            status.m_PendingSites.Add(MapFromProtocolString(sField));
            int nValue = atoi(sValue);
            status.m_PendingEntries.Add(nValue);
          }
        }
        else
          bCurrentlyParsingSites = FALSE;
      }
      
      if (sLine.Find("Database entries by category:") == 0)
        bCurrentlyParsingCategories = TRUE;
      else if (sLine.Find("Pending File transmissions:") == 0)
        bCurrentlyParsingSites = TRUE;
      else if (SplitSemiColonLine(sLine, sField, sValue))
      {
        if (sField == "current proto")
          status.m_nCurrentProtocol = atoi(sValue);  
        else if (sField == "max proto")
          status.m_nMaxProtocol = atoi(sValue);
        else if (sField == "gets")
        {
          if (sValue == "yes")
            status.m_bGetsAllowed = TRUE;
          else if (sValue == "no")
            status.m_bGetsAllowed = FALSE;
          else
          {
  		      TRACE(_T("CCDDB::ParseStatusBody, CDDB server failed to return a valid gets response to the stat command\n"));
  		      return FALSE;
          }
        }
        else if (sField == "updates")
        {
          if (sValue == "yes")
            status.m_bUpdatesAllowed = TRUE;
          else if (sValue == "no")
            status.m_bUpdatesAllowed = FALSE;
          else
          {
  		      TRACE(_T("CCDDB::ParseStatusBody, CDDB server failed to return a valid updates response to the stat command\n"));
  		      return FALSE;
          }
        }
        else if (sField == "posting")
        {
          if (sValue == "yes")
            status.m_bPostingAllowed = TRUE;
          else if (sValue == "no")
            status.m_bPostingAllowed = FALSE;
          else
          {
  		      TRACE(_T("CCDDB::ParseStatusBody, CDDB server failed to return a valid posting response to the stat command\n"));
  		      return FALSE;
          }
        }
        else if (sField == "quotes")
        {
          if (sValue == "yes")
            status.m_bQuotes = TRUE;
          else if (sValue == "no")
            status.m_bQuotes = FALSE;
          else
          {
  		      TRACE(_T("CCDDB::ParseStatusBody, CDDB server failed to return a valid quotes response to the stat command\n"));
  		      return FALSE;
          }
        }
        else if (sField == "current users")
          status.m_nCurrentUsers = atoi(sValue);  
        else if (sField == "max users")
          status.m_nMaxUsers = atoi(sValue);  
        else if (sField == "strip ext")
        {
          if (sValue == "yes")
            status.m_bStripExtended = TRUE;
          else if (sValue == "no")
            status.m_bStripExtended = FALSE;
          else
          {
  		      TRACE(_T("CCDDB::ParseStatusBody, CDDB server failed to return a valid strip extended response to the stat command\n"));
  		      return FALSE;
          }
        }
        else if (sField == "Database entries")
          status.m_nDatabaseEntries = atoi(sValue);  
      }
    }

    //Everything went ok, so set the return value
    bSuccess = TRUE;
  }

  return bSuccess;
}

void CCDDB::Query(const CCDDBSite& server, DWORD dwDiscID, const CArray<CCDDBTrackPosition, CCDDBTrackPosition&>& tracks, 
                  CArray<CCDDBQueryResult, CCDDBQueryResult&>& results)
{
  //Remove any entries which are in the results array
  results.RemoveAll();

  CWSocket socket;
  try
  {
    //Create the socket
    socket.Create();

    //Connect to the host on the specified port
    BOOL bUseProxy = !m_sProxyHost.IsEmpty();
    socket.Connect(bUseProxy ? m_sProxyHost : server.m_sSite, bUseProxy ? m_nProxyPort : server.m_nPort);

    //Form the specialized HTTP request to retrieve the query data
    CString sDiscID;
    sDiscID.Format(_T("%08x"), dwDiscID);

    int nTracks = static_cast<int>(tracks.GetSize()) - 1; //Number of tracks is 1 less than the size of the array as it also contains the lead out position
    CString sTracks;
    sTracks.Format(_T("%d"), nTracks);

    CString sOffsets;
    for (int i=0; i<nTracks; i++)
    {
      //Form the frame number
      CString sFrame;
      sFrame.Format(_T("%d"), tracks[i].m_nMinute*60*75 + tracks[i].m_nSecond*75 + tracks[i].m_nFrame);

      //Accumulate the frame number into the string
      sOffsets += sFrame;

      //Add a space between each offset
      if (i<(nTracks-1))
       sOffsets += _T("+");
    }

    //Compute the playing length in seconds by using the lead out position
    int nCDlength = tracks[nTracks].m_nMinute*60 + tracks[nTracks].m_nSecond;
    CString sLength;
    sLength.Format(_T("%d"), nCDlength);

    //Form the full request string
    CString sRequest;
    if (!bUseProxy)
      sRequest.Format(_T("GET %s?cmd=cddb+query+%s+%s+%s+%s&%s HTTP/1.0\r\n\r\n"), server.m_sAddress.operator LPCTSTR(), sDiscID.operator LPCTSTR(), sTracks.operator LPCTSTR(), 
                      sOffsets.operator LPCTSTR(), sLength.operator LPCTSTR(), GetHelloCommand().operator LPCTSTR());
    else
      sRequest.Format(_T("GET http://%s%s?cmd=cddb+query+%s+%s+%s+%s&%s HTTP/1.0\r\n%s\r\n\r\n"), server.m_sSite.operator LPCTSTR(), server.m_sAddress.operator LPCTSTR(), sDiscID.operator LPCTSTR(), 
                      sTracks.operator LPCTSTR(), sOffsets.operator LPCTSTR(), sLength.operator LPCTSTR(), GetHelloCommand().operator LPCTSTR(), GetProxyHeader(server.m_sSite).operator LPCTSTR());

    //Send the request through the socket
    CT2A szRequest(sRequest);
    int nCmdLength = static_cast<int>(strlen(szRequest));
    socket.Send(szRequest, nCmdLength);
    
    //Read the response back through the socket
    CStringA sResponse;
    sResponse.Preallocate(4096);
    if (ReadResponse(socket, "\r\n.\r\n", sResponse, ReadResponseQuery))
    {
      //Extract the HTTP body from the response;
      LPCSTR pszBody = FindHTTPBody(sResponse);

      //Parse out the query info from the body
      if (!ParseQueryBody(pszBody, sDiscID, results))
        ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_QUERY_RESPONSE, m_sLastCommandResponse);
    }
    else
      ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_QUERY_RESPONSE, m_sLastCommandResponse);
  }
  catch(CWSocketException* pEx)
  {
    DWORD dwError = pEx->m_nError;
    pEx->Delete();
    
    ThrowCDDBException(dwError, FACILITY_WIN32);
  }
}

BOOL CCDDB::ParseQueryBody(LPCSTR pszBody, const CString& sDiscID, CArray<CCDDBQueryResult, CCDDBQueryResult&>& results)
{
#ifndef _DEBUG
  sDiscID;   //To get rid of level 4 warning in release mode
#endif

  //What will be the return value from this function (assume the worst)
  BOOL bSuccess = FALSE;

  //From the HTTP body get the CDDB response code
  int nResponseCode = GetCDDBReponseCode(pszBody);

  //If the response is in the expected range (21x) then parse the query data line by line
  if (nResponseCode >= 210 && nResponseCode <= 219)
  {
    //Skip over the current line
    LPCSTR pszLine = pszBody;
    CStringA sLine;
    GetNextLine(pszLine, sLine);

    //Parse each line
    while (pszLine)
    {
      //parse the current line and get the next line
      GetNextLine(pszLine, sLine);

      //If we can parse the current line then add it 
      //to the query array
      CCDDBQueryResult result;
      if (ParseQueryBodyHelper(result, sLine))
      {
        results.Add(result);
        bSuccess = TRUE;
      }
    }
  }
  else if (nResponseCode == 200 && (strlen(pszBody) >= 5))
  {
    //found an exact match, If we can parse the 
    //current line then add it to the query array
    CCDDBQueryResult result;
    if (ParseQueryBodyHelper(result, pszBody+4))
    {
      results.Add(result);
      bSuccess = TRUE;
    }
  }
  else if (nResponseCode == 202)
  {
    //Deliberately do nothing other than a trace
    TRACE(_T("CCDDB::ParseQueryBody, CDDB server does not contain the specifed album with DISCID:%s in its database\n"), sDiscID.operator LPCTSTR());
  }

  return bSuccess;
}

void CCDDB::Read(const CCDDBSite& server, DWORD dwDiscID, const CString& sCategory, CCDDBRecord& record)
{
  CWSocket socket;
  try
  {
    //Create the socket
    socket.Create();

    //Connect to the host on the specified port
    BOOL bUseProxy = !m_sProxyHost.IsEmpty();
    socket.Connect(bUseProxy ? m_sProxyHost : server.m_sSite, bUseProxy ? m_nProxyPort : server.m_nPort);

    //Form the full request string
    CString sDiscID;
    sDiscID.Format(_T("%08x"), dwDiscID);

    CString sRequest;
    if (!bUseProxy)
      sRequest.Format(_T("GET %s?cmd=cddb+read+%s+%s&%s HTTP/1.0\r\n\r\n"), server.m_sAddress.operator LPCTSTR(), sCategory.operator LPCTSTR(), sDiscID.operator LPCTSTR(), GetHelloCommand().operator LPCTSTR());
    else
      sRequest.Format(_T("GET http://%s%s?cmd=cddb+read+%s+%s&%s HTTP/1.0\r\n%s\r\n\r\n"), server.m_sSite.operator LPCTSTR(), server.m_sAddress.operator LPCTSTR(), sCategory.operator LPCTSTR(), 
                      sDiscID.operator LPCTSTR(), GetHelloCommand().operator LPCTSTR(), GetProxyHeader(server.m_sSite).operator LPCTSTR());  

    //Send the request through the socket
    CT2A szRequest(sRequest);
    int nCmdLength = static_cast<int>(strlen(szRequest));
    socket.Send(szRequest, nCmdLength);

    //Read the response back through the socket
    CStringA sResponse;
    sResponse.Preallocate(4096);
    if (ReadResponse(socket, "\r\n.\r\n", sResponse))
    {
      //Extract the HTTP body from the response;
      LPCSTR pszBody = FindHTTPBody(sResponse);

      //Parse the CDDB record
      if (!ParseRecordBody(pszBody, record))
        ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_READ_RESPONSE, m_sLastCommandResponse);

      if (record.m_dwDiscID != dwDiscID)
      {
        TRACE(_T("CCDDB::Read, Returned DiscID from CDDB read (%08x) is different than the computed value (%08x), using the computed value\n"), record.m_dwDiscID, dwDiscID);
        record.m_dwDiscID = dwDiscID;      
      }
    }
    else
      ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_QUERY_RESPONSE, m_sLastCommandResponse);
  }
  catch(CWSocketException* pEx)
  {
    DWORD dwError = pEx->m_nError;
    pEx->Delete();
    
    ThrowCDDBException(dwError, FACILITY_WIN32);
  }
}

BOOL CCDDB::ParseRecordBody(LPCSTR pszBody, CCDDBRecord& record)
{
  //From the HTTP body get the CDDB response code
  int nResponseCode = GetCDDBReponseCode(pszBody);

  //If the response is in the expected range (21x) then parse the query data line by line
  if (nResponseCode >= 210 && nResponseCode <= 219)
  {
    //Skip over the current line
    LPCSTR pszLine = pszBody;
    CStringA sLine;
    GetNextLine(pszLine, sLine);
  
    //Hand of to the helper function
    if (!ParseRecordBodyHelper(record, pszLine))
      return FALSE;
  }

  //Return the status code
  return TRUE;
}

void CCDDB::PrepareTextForSubmit(CString& sText)
{
  sText.Replace(_T("\\"), _T("\\\\"));
  sText.Replace(_T("\n"), _T("\\n"));
  sText.Replace(_T("\t"), _T("\\t"));
}

CString CCDDB::FormSubmitLine(const CString& sKeyword, const CString& sValue)
{
  CString sLine;
  CString sLocalValue(sValue);
  PrepareTextForSubmit(sLocalValue);

  do
  {
    int nTotalSize = sKeyword.GetLength() + sLocalValue.GetLength() + 1;
    int nLineLength = min(254, nTotalSize);
    int nValueLength = nLineLength - sKeyword.GetLength() - 1;

    //Form the current line
    CString sBuf;
    sBuf.Format(_T("%s=%s\r\n"), sKeyword.operator LPCTSTR(), sLocalValue.Left(nValueLength).operator LPCTSTR());

    //Add to what will be the return value
    sLine += sBuf;

    //Prepare for the next time around
    sLocalValue = sLocalValue.Right(sLocalValue.GetLength() - nValueLength);
  }
  while (sLocalValue.GetLength());

  return sLine;
}


void CCDDB::Submit(const CCDDBSite& server, const CString& sCategory, const CString& sEmailAddress, 
                   CCDDBRecord& record, const CString& sRejectionNotice, BOOL bReallySubmit)
{
  CWSocket socket;
  try
  {
    //Create the socket
    socket.Create();

    //Connect to the host on the specified port
    BOOL bUseProxy = !m_sProxyHost.IsEmpty();
    socket.Connect(bUseProxy ? m_sProxyHost : server.m_sSite, bUseProxy ? m_nProxyPort : server.m_nPort);
    
    //Form the specialized HTTP request to submit the album

    //Form the body
    
    //First the first few intro lines
    CString sBody;
    sBody.Preallocate(4096);
    sBody = _T("# xmcd\r\n#\r\n# Track frame offsets:\r\n");

    //Then the track offsets
    CString sBuf;
    for (INT_PTR i=0; i<record.m_TrackOffsets.GetSize(); i++) 
    {
      sBuf.Format(_T("# %d\r\n"), record.m_TrackOffsets.GetAt(i));
      sBody += sBuf;
    }

    //Then the disc length
    sBuf.Format(_T("#\r\n# Disc length: %d seconds\r\n#\r\n"), record.m_nDiskLength); 
    sBody += sBuf;

    //Then the database revision
    sBuf.Format(_T("#\r\n# Revision: %d\r\n"), record.m_nDatabaseRevision); 
    sBody += sBuf;

    //Then the client details
    sBuf.Format(_T("# Submitted via: %s %s %s"), record.m_sClientName.operator LPCTSTR(), record.m_sClientVersion.operator LPCTSTR(), record.m_sClientComments.operator LPCTSTR());
    PrepareTextForSubmit(sBuf);
    ASSERT(sBuf.GetLength() <= 254); //Only 254 chars per line supported for the "# Submitted via:" line
    sBody += sBuf; 
    sBody += _T("\r\n#\r\n");

    sBuf.Format(_T("DISCID=%08x\r\n"), record.m_dwDiscID);  
    sBody += sBuf; //Then the discid

    //Then the DTITLE
    if (record.m_sArtist.GetLength())
      sBuf.Format(_T("%s / %s"), record.m_sArtist.operator LPCTSTR(), record.m_sTitle.operator LPCTSTR());
    else
      sBuf = record.m_sTitle;
    sBuf = FormSubmitLine(_T("DTITLE"), sBuf);
    sBody += sBuf;  

    if (m_nProtocolVersion >= 5)
    {
      //Then the DYEAR
      ASSERT(record.m_nYear >= 0 && record.m_nYear <= 9999);
      if (record.m_nYear != 0)
        sBuf.Format(_T("DYEAR=%04d\r\n"), record.m_nYear);
      else
        sBuf = _T("DYEAR=\r\n");
      sBody += sBuf;

      //Then the DGENRE
      sBuf = FormSubmitLine(_T("DGENRE"), record.m_sGenre);
      sBody += sBuf;
    }

    //Then the track titles
    for (INT_PTR i=0; i<record.m_TrackTitles.GetSize(); i++)  
    {
      CString sKeyword;
      sKeyword.Format(_T("TTITLE%d"), i);
      sBuf = FormSubmitLine(sKeyword, record.m_TrackTitles.ElementAt(i));
      sBody += sBuf;
    }

    //Then the extended data
    sBuf = FormSubmitLine(_T("EXTD"), record.m_ExtendedData);
    sBody += sBuf; 

    //Then the extended track data
    for (INT_PTR i=0; i<record.m_ExtendedTrackData.GetSize(); i++) 
    {
      CString sKeyword;
      sKeyword.Format(_T("EXTT%d"), i);
      sBuf = FormSubmitLine(sKeyword, record.m_ExtendedTrackData.ElementAt(i));
      sBody += sBuf;
    }

    //Finally the PLAYORDER line (which is deliberately empty)
    sBody += _T("PLAYORDER="); 

    //Form the header into "sRequest"
    CString sRequest;
    sRequest.Preallocate(4096);
    if (!bUseProxy)
      sRequest.Format(_T("POST %s HTTP/1.0\r\nCategory: "), server.m_sAddress.operator LPCTSTR());
    else
      sRequest.Format(_T("POST http://%s%s HTTP/1.0\r\n%s\r\nCategory: "), server.m_sSite.operator LPCTSTR(), server.m_sAddress.operator LPCTSTR(), 
                      GetProxyHeader(server.m_sSite).operator LPCTSTR());

    sRequest += sCategory; //Then the category
    sRequest += _T("\r\n");

    sBuf.Format(_T("Discid: %08x\r\n"), record.m_dwDiscID); //Then the DISCID
    sRequest += sBuf;

    sBuf.Format(_T("User-Email: %s\r\n"), sEmailAddress.operator LPCTSTR());
    sRequest += sBuf; //Then the User's email address

    if (bReallySubmit) //Then the submit mode
      sRequest += _T("Submit-Mode: submit\r\n");
    else
      sRequest += _T("Submit-Mode: test\r\n");                        
    
  #ifdef _UNICODE
    if (m_nProtocolVersion >= 6)
      sRequest += _T("Charset: UTF-8\r\n");
    else
      sRequest += _T("Charset: ISO-8859-1\r\n");
  #else
    sRequest += _T("Charset: ISO-8859-1\r\n");
  #endif    

    ASSERT(sRejectionNotice.GetLength() <= 70); //Rejection notice is limited to 70 characters
    sBuf.Format(_T("X-Cddbd-Note: %s\r\n"), sRejectionNotice.operator LPCTSTR()); //Then the rejection notice
    sRequest += sBuf;

    sBuf.Format(_T("Content-Length: %d\r\n\r\n"), sBody.GetLength()); 
    sRequest += sBuf; //Then the Content-Length

    //Add the body to the end of the request
    sRequest += sBody;   

    //Send the request through the socket
    CStringA sASCIIRequest(MapToProtocolString(sRequest));
    socket.Send(sASCIIRequest, sASCIIRequest.GetLength());

    //Read the response back through the socket
    CStringA sResponse;
    sResponse.Preallocate(4096);
    if (ReadResponse(socket, "\r\n.\r\n", sResponse, ReadResponseSubmit))
    {
      //Extract the HTTP body from the response;
      LPCSTR pszBody = FindHTTPBody(sResponse);

      //Parse out the response
      if (!ParseSubmitBody(pszBody))
        ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_SUBMIT_RESPONSE, m_sLastCommandResponse);
    }
    else
      ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_SUBMIT_RESPONSE, m_sLastCommandResponse);
  }
  catch(CWSocketException* pEx)
  {
    DWORD dwError = pEx->m_nError;
    pEx->Delete();
    
    ThrowCDDBException(dwError, FACILITY_WIN32);
  }
}

BOOL CCDDB::ParseSubmitBody(LPCSTR pszBody)
{
  //From the HTTP body get the CDDB response code
  int nResponseCode = GetCDDBReponseCode(pszBody);

  //If the response is in the expected range (20x) then return success
  return (nResponseCode >= 200 && nResponseCode <= 209);
}

void CCDDB::MessageOfTheDay(CString& sMessage, const CString& sServer, const CString& sAddress, int nPort)
{
  //Form the site which we will be getting the message from
  CCDDBSite server;
  server.m_sSite = sServer;
  server.m_nPort = nPort;
  server.m_sAddress = sAddress;
  
  //Let the other version 
  MessageOfTheDay(server, sMessage);
}

void CCDDB::MessageOfTheDay(const CCDDBSite& server, CString& sMessage)
{
  CWSocket socket;
  try
  {
    //Create the socket
    socket.Create();

    //Connect to the host on the specified port
    BOOL bUseProxy = !m_sProxyHost.IsEmpty();
    socket.Connect(bUseProxy ? m_sProxyHost : server.m_sSite, bUseProxy ? m_nProxyPort : server.m_nPort);

    //Form the specialized HTTP request to retrieve the status data
    CString sRequest;
    if (!bUseProxy)
      sRequest.Format(_T("GET %s?cmd=motd&%s HTTP/1.0\r\n\r\n"), server.m_sAddress.operator LPCTSTR(), GetHelloCommand().operator LPCTSTR());
    else
      sRequest.Format(_T("GET http://%s%s?cmd=motd&%s HTTP/1.0\r\n%s\r\n\r\n"), server.m_sSite.operator LPCTSTR(), server.m_sAddress.operator LPCTSTR(), 
                      GetHelloCommand().operator LPCTSTR(), GetProxyHeader(server.m_sSite).operator LPCTSTR());

    //Send the request through the socket
    CT2A szRequest(sRequest);
    int nCmdLength = static_cast<int>(strlen(szRequest));
    socket.Send(szRequest, nCmdLength);

    //Read the response back through the socket
    CStringA sResponse;
    sResponse.Preallocate(4096);
    if (ReadResponse(socket, "\r\n.\r\n", sResponse))
    {
      //Extract the HTTP body from the response;
      LPCSTR pszBody = FindHTTPBody(sResponse);

      //Parse out the message from the body
      if (!ParseMessageBody(pszBody, sMessage))
        ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_MOTD_RESPONSE, m_sLastCommandResponse);
    }
    else
      ThrowCDDBException(IDS_MFCCDDB_UNEXPECTED_MOTD_RESPONSE, m_sLastCommandResponse);
  }
  catch(CWSocketException* pEx)
  {
    DWORD dwError = pEx->m_nError;
    pEx->Delete();
    
    ThrowCDDBException(dwError, FACILITY_WIN32);
  }
}

BOOL CCDDB::ParseMessageBody(LPCSTR pszBody, CString& sMessage)
{
  //What will be the return value from this function (assume the worst)
  BOOL bSuccess = FALSE;

  //From the HTTP body get the CDDB response code
  int nResponseCode = GetCDDBReponseCode(pszBody);

  //If the response is in the expected range (21x) then copy the message into "sMessage"
  if (nResponseCode >= 210 && nResponseCode <= 219)
  {
    //Skip over the current line
    LPCSTR pszLine = pszBody;
    CStringA sLine;
    GetNextLine(pszLine, sLine);
  
    //Parse each line
    sMessage.Empty();
    while (pszLine)
    {
      //parse the current line and get the next line
      GetNextLine(pszLine, sLine);

      if (sLine != ".")
      {
        sMessage += sLine;
        sMessage += "\r\n";
      }
      
      bSuccess = TRUE;
    }
  }

  return bSuccess;
}

void CCDDB::SetProxyDetails(const CString& sHost, int nPort, const CString& sUser, const CString& sPassword)
{
  m_sProxyHost = sHost;
  m_nProxyPort = nPort;
  m_sProxyUser = sUser;
  m_sProxyPassword = sPassword;
}

CString CCDDB::GetProxyHeader(const CString& sHost)
{
  //Validate the parameters
  ASSERT(m_sProxyHost.GetLength());

  //What will be the return value from this function
  CString sLine;

  //If we need authentication
  if (m_sProxyUser.GetLength())
  {
    CString sCombinedUserAndPassword;
    sCombinedUserAndPassword.Format(_T("%s:%s"),m_sProxyUser.operator LPCTSTR(), m_sProxyPassword.operator LPCTSTR());
	  CT2A szAsciiCombinedUserAndPassword(sCombinedUserAndPassword);
    int nSize = static_cast<int>(strlen(szAsciiCombinedUserAndPassword));

    //allocate the encoded buffer
    int nOutSize = ATL::Base64EncodeGetRequiredLength(nSize, ATL_BASE64_FLAG_NOPAD | ATL_BASE64_FLAG_NOCRLF);
    char* pszEncoded = new char[nOutSize];

    //Do the encoding
    ATL::Base64Encode(reinterpret_cast<const BYTE*>(szAsciiCombinedUserAndPassword.operator LPSTR()), nSize, pszEncoded, &nOutSize, ATL_BASE64_FLAG_NOPAD | ATL_BASE64_FLAG_NOCRLF);

    //Form the header 
    sLine.Format(_T("Proxy-Authorization: Basic %s\r\nHost: %s"), CA2T(pszEncoded).operator LPTSTR(), sHost.operator LPCTSTR());

    //delete the encoded temp buffer
    delete [] pszEncoded;
  }
  else
  {
    //Just add the Host header
    sLine.Format(_T("Host: %s"), sHost.operator LPCTSTR());
  }

  //Return the formed string
  return sLine;
}

BOOL CCDDB::Read(const CString& sDirectory, DWORD dwDiscID, BOOL bStandardForm, CCDDBRecord& record, CString& sFileSystemGenre)
{
  //What will be the return value from this function (assume the worst)
  BOOL bFound = FALSE;

  //First iterate thro the genre subdirectories
  CFileFind genreFileFind;
  BOOL bContinueGenreSearch = genreFileFind.FindFile(sDirectory + _T("\\*.*"));
  while (bContinueGenreSearch && !bFound)
  {
    //Get ready for the next time around
    bContinueGenreSearch = genreFileFind.FindNextFile();
    if (!genreFileFind.IsDots() && genreFileFind.IsDirectory())
    {
      if (bStandardForm)
      {
        CString sFilename;
        sFilename.Format(_T("%s\\%08x"), genreFileFind.GetFilePath().operator LPCTSTR(), dwDiscID);
        
        try
        {
          ParseRecordBodyFromFile(dwDiscID, sFilename, TRUE, record);
          bFound = TRUE;
        }
        catch(CCDDBException* pEx)
        {
          pEx->Delete();
        }

        //Update the Genre details of the record
        if (bFound)
          sFileSystemGenre = genreFileFind.GetFileTitle();
      }
      else
      {
        //Now iterate thro all the files in this genre directory
        CFileFind filesFileFind;
        BOOL bContinueFileSearch = filesFileFind.FindFile(genreFileFind.GetFilePath() + _T("\\??to??"));
        while (bContinueFileSearch && !bFound)
        {
          //Get ready for the next time around
          bContinueFileSearch = filesFileFind.FindNextFile();

          //Is the filename in the range of the discid we are looking for
          CString sFile(filesFileFind.GetFileTitle());

          DWORD dwLowerRange = _tcstoul(sFile.Left(2), NULL, 16);
          DWORD dwUpperRange = _tcstoul(sFile.Mid(4, 2), NULL, 16);
          DWORD dwTruncatedDiscID = dwDiscID>>24;
          if ((dwLowerRange <= dwTruncatedDiscID) && (dwTruncatedDiscID <= dwUpperRange))
          {
            try
            {
              ParseRecordBodyFromFile(dwDiscID, filesFileFind.GetFilePath(), FALSE, record);
              bFound = TRUE;
            }
            catch(CCDDBException* pEx)
            {
              pEx->Delete();
            }

            //Update the Genre details of the record
            if (bFound)
              sFileSystemGenre = genreFileFind.GetFileTitle();
          }
        }
      }
    }
  }

  return bFound;
}

void CCDDB::ParseRecordBodyFromFile(DWORD dwDiscID, const CString& sFile, BOOL bStandardForm, CCDDBRecord& record)
{
  if (bStandardForm)
  {
    ATL::CHandle file(CreateFile(sFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, NULL));
    if (file != INVALID_HANDLE_VALUE)
    {
      //Get the length of the file (we only support parsing files less than 4GB when in standard form)
      DWORD dwFileSizeHigh = 0;
      DWORD dwFileSizeLow = GetFileSize(file, &dwFileSizeHigh);
      DWORD dwError = GetLastError();
      if (dwFileSizeLow == INVALID_FILE_SIZE)
        ThrowCDDBException(dwError, FACILITY_WIN32);
      if (dwFileSizeHigh)
        ThrowCDDBException(IDS_MFCCDDB_FILE_SIZE_TO_PARSE_TOO_LARGE, FACILITY_ITF);
      if (dwFileSizeLow == 0)
        ThrowCDDBException(IDS_MFCCDDB_CANNOT_PARSE_ZERO_BYTE_RECORD, FACILITY_ITF);
      
      //Read the whole record into memory
      char* pszRecordData = new char[dwFileSizeLow+1];
      pszRecordData[0] = '\0';
      DWORD dwRead = 0;
      if (ReadFile(file, pszRecordData, dwFileSizeLow, &dwRead, NULL))
        pszRecordData[dwRead] = '\0';

      //Hand of all the work to the helper function
      BOOL bParsed = ParseRecordBodyHelper(record, pszRecordData);

      //tidy up the heap memory we have used
      delete [] pszRecordData;

      if (!bParsed)
        ThrowCDDBException(IDS_MFCCDDB_CANNOT_PARSE_RECORD, FACILITY_ITF);
    }
    else
      ThrowCDDBException();
  }
  else
  {
    CStdioFile file;
    CFileException fileException;
    if (file.Open(sFile, CFile::typeText | CFile::modeRead | CFile::shareDenyWrite, &fileException))
    {
      CString sRecordData;
      sRecordData.Preallocate(4096);

      CString sFirstInterestingLine;
      sFirstInterestingLine.Format(_T("#FILENAME=%08X"), dwDiscID);

      BOOL bMore = TRUE;
      BOOL bFound = FALSE;
      while (bMore)
      {
        CString sLine;
        try
        {
          bMore = file.ReadString(sLine);
          if (sLine.CompareNoCase(sFirstInterestingLine) == 0)
            bFound = TRUE;
          else if (bFound)
          {
            if (sLine.Find(_T("#FILENAME")) != -1)
              bMore = FALSE;
            else
            {
              sRecordData += sLine;
              sRecordData += _T("\r\n");
            }
          }
        }
        catch(CFileException* pEx)
        {
          DWORD dwError = pEx->m_lOsError;
          pEx->Delete();
          ThrowCDDBException(dwError);
        }
      }

      //Hand of all the work to the helper function
      if (bFound)
      {
        if (!ParseRecordBodyHelper(record, CStringA(sRecordData)))
          ThrowCDDBException(IDS_MFCCDDB_CANNOT_PARSE_RECORD, FACILITY_ITF);
      }  
      else
        ThrowCDDBException(IDS_MFCCDDB_CANNOT_FIND_RECORD, FACILITY_ITF);
    }
    else
      ThrowCDDBException(fileException.m_lOsError);
  }
}

CString CCDDB::MapFromProtocolString(const CStringA& sInput)
{
  //What will be the return value from this function
  CString sOutput;

#ifdef _UNICODE
  if (m_nProtocolVersion >= 6)
  {
    int nInputLen = sInput.GetLength();
    int nChars = MultiByteToWideChar(CP_UTF8, 0, sInput, nInputLen, NULL, 0); 
    if (nChars)
    {
	    ATL::CHeapPtr<wchar_t> szUnicode;
	    if (!szUnicode.Allocate(nChars + 1))
	      CCDDB::ThrowCDDBException(ERROR_OUTOFMEMORY, FACILITY_WIN32); 
      nChars = MultiByteToWideChar(CP_UTF8, 0, sInput, nInputLen, szUnicode, nChars);
      if (nChars)
      {
        szUnicode[nChars] = '\0';
        sOutput = szUnicode;
      }
    }
  }
  else
    sOutput = CA2T(sInput); //Do a simple conversion from ASCII to Unicode
#else
  sOutput = CA2T(sInput); //Do a simple conversion from ASCII to Unicode
#endif  

  return sOutput;
}

CStringA CCDDB::MapToProtocolString(const CString& sInput)
{
  //What will be the return value from this function
  CStringA sOutput;

#ifdef _UNICODE
  if (m_nProtocolVersion >= 6)
  {
    int nInputLen = sInput.GetLength();
    int nChars = WideCharToMultiByte(CP_UTF8, 0, sInput, nInputLen, NULL, 0, NULL, NULL); 
    if (nChars)
    {
      ATL::CHeapPtr<char> szUTF8;
      if (!szUTF8.Allocate(nChars + 1))
        CCDDB::ThrowCDDBException(ERROR_OUTOFMEMORY, FACILITY_WIN32); 
      nChars = WideCharToMultiByte(CP_UTF8, 0, sInput, nInputLen, szUTF8, nChars, NULL, NULL);
      if (nChars)
      {
        szUTF8[nChars] = '\0';
        sOutput = szUTF8;
      }
    }
  }
  else
    sOutput = CT2A(sInput); //Do a simple conversion from Unicode to ASCII
#else
  sOutput = CT2A(sInput); //Do a simple conversion from Unicode to ASCII
#endif

  return sOutput;
}

void CCDDB::GetNextLine(LPCSTR& pszLine, CStringA& sLine)
{
  if (pszLine)
  {
    //Find the start of the next line. Both DOS/Windows (\r\n) 
    //and UNIX (\n) terminators are valid. First try to
    //find a DOS EOL
    const char* pszNextLine = strstr(pszLine, "\r\n");
    if (pszNextLine)
    {
      sLine.SetString(pszLine, static_cast<int>(pszNextLine - pszLine));
      pszLine = pszNextLine + 2;
    }
    else
    {
      //Failed to get a DOS/Windows EOL, try to get a UNIX EOL instead
      pszNextLine = strstr(pszLine, "\n");
      if (pszNextLine)
      {
        sLine.SetString(pszLine, static_cast<int>(pszNextLine - pszLine));
        pszLine = pszNextLine + 1;
      }
      else
        pszLine = NULL;
    }
  }
}

BOOL CCDDB::ParseKeywordLine(const CStringA& sLine, CStringA& sKeyword, CStringA& sValue)
{
  //what will be the return value from this function (assume the worst)
  BOOL bSuccess = FALSE;
  
  int nSeparator = sLine.Find("=");
  if (nSeparator != -1)
  {
    bSuccess = TRUE;
    
    sKeyword = sLine.Left(nSeparator);
    sValue = sLine.Right(sLine.GetLength() - nSeparator - 1);

    //Replace all strings as specified in the cddb document
    sValue.Replace("\\n", "\n");
    sValue.Replace("\\t", "\t");
    sValue.Replace("\\\\", "\\");
  }

  return bSuccess;
}

BOOL CCDDB::ParseCommentLine(const CStringA& sLine, CStringA& sValue)
{
  //What will be the return value from the function (assume the worst)
  BOOL bSuccess = FALSE;
  
  //Do the actual parsing
  if (sLine.Left(1) == _T('#'))
  {
    bSuccess = TRUE;
    sValue = sLine.Right(sLine.GetLength() - 1);
    sValue.Trim();
  }

  return bSuccess;
}

BOOL CCDDB::ParseRecordBodyHelper(CCDDBRecord& record, const CStringA& sRecord)
{
  //Setup all the booleans which indicate a successful record parse.
  //These will be checked at the end of the function
  BOOL bFoundXMCDComment = FALSE;
  BOOL bParsingTrackFrameOffsets = FALSE;
  BOOL bFoundDiscLength = FALSE;
  BOOL bFoundSubmittedVia = FALSE;
  BOOL bFoundDISCID = FALSE;
  BOOL bFoundDTITLE = FALSE;
  CString sTitle;
  CString sPlayOrder;

  //initialize to default values
  record = CCDDBRecord();

  //Loop thro each line 
  LPCSTR pszCurLine = sRecord;
  while (pszCurLine)
  {
    //Get the next line before we parse the current one
    CStringA sLine;
    GetNextLine(pszCurLine, sLine);

    //Is it a comment line?
    CStringA sValue;
    if (ParseCommentLine(sLine, sValue))
    {
      if (sValue.Find("xmcd") == 0)
      {
        //Handle the "# xmcd" line
        bFoundXMCDComment = TRUE;
      }
      else if (sValue.Find("Track frame offsets:") != -1)
      {
        //Handle the "# Track frame offsets:" line
        bParsingTrackFrameOffsets = TRUE;
      }
      else if (sValue.Find("Disc length:") == 0)
      {
        //If we've found the disc length line then there are no more track frame offset lines
        bParsingTrackFrameOffsets = FALSE;

        //Handle the "# Disc length:" line
        bFoundDiscLength = TRUE;
        sValue = sValue.Right(sValue.GetLength() - 13);
        record.m_nDiskLength = atoi(sValue);
      }
      else if (sValue.Find("Revision:") == 0)
      {
        //Handle the "# Revision:" line
        sValue = sValue.Right(sValue.GetLength() - 9);
        record.m_nDatabaseRevision = atoi(sValue);
      }
      else if (sValue.Find("Submitted via:") == 0)
      {
        //Handle the "# Submitted via:" line
        bFoundSubmittedVia = TRUE;

        static char* pszSeparators = " ";
        sValue = sValue.Right(sValue.GetLength() - 14);
        char* pszValue = sValue.GetBuffer(sValue.GetLength());
        char* pszNextToken = NULL;
        LPSTR pszToken = strtok_s(pszValue, pszSeparators, &pszNextToken);
        if (pszToken)
          record.m_sClientName = MapFromProtocolString(pszToken);
        else
        {
          sValue.ReleaseBuffer();
          TRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a valid Submitted via: line\n"));
          return FALSE;
        }

        //Pull out the client version
        pszToken = strtok_s(NULL, pszSeparators, &pszNextToken);  
        if (pszToken)
          record.m_sClientVersion = MapFromProtocolString(pszToken);
        else
        {
          sValue.ReleaseBuffer();
          TRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a valid Submitted via: line\n"));
          return FALSE;
        }

        //Pull out the client comments
        pszToken = strtok_s(NULL, pszSeparators, &pszNextToken);    
        if (pszToken)
          record.m_sClientComments = MapFromProtocolString(pszToken);
        else
          record.m_sClientComments.Empty();

        sValue.ReleaseBuffer();
      }
      else if (bParsingTrackFrameOffsets)
      {
        if (sValue.GetLength())
        {
          //Handle all the actual track frame offsets lines
          int nTrackOffset = atoi(sValue);
          record.m_TrackOffsets.Add(nTrackOffset);
        }
      }
    }
    else
    {
      //Musn't be a comment line
      CStringA sKeyword;
      if (ParseKeywordLine(sLine, sKeyword, sValue))
      {
        if (sKeyword == "DISCID")
        {
          if (sValue.GetLength() >= 8)
          {
            //handle the "DISCID" line
            bFoundDISCID = TRUE;
            sscanf_s(sValue, "%x", &record.m_dwDiscID);
          }            
        }
        else if (sKeyword == "DTITLE")
        {
          //Handle the "DTITLE" line
          bFoundDTITLE = TRUE;
          sTitle += MapFromProtocolString(sValue);
        }
        else if (sKeyword == "DYEAR")
        {
          //Handle the "DYEAR" line
          record.m_nYear = atoi(sValue);
        }
        else if (sKeyword == "DGENRE")
        {
          //Handle the "DGENRE" line
          record.m_sGenre += MapFromProtocolString(sValue);
        }
        else if (sKeyword.Find("TTITLE") == 0)
        {
          //Handle the "TTITLE" lines

          //work out the offset and add the string
          CStringA sTrack(sKeyword.Right(sKeyword.GetLength() - 6));
          int nOffset = atoi(sTrack);
          record.m_TrackTitles.SetSize(nOffset+1);
          record.m_TrackTitles.ElementAt(nOffset) += MapFromProtocolString(sValue);
        }
        else if (sKeyword == "EXTD")
        {
          //Handle any "EXTD" lines
          record.m_ExtendedData += MapFromProtocolString(sValue);
        }
        else if (sKeyword.Find("EXTT") == 0)
        {
          //work out the offset and add the string
          CStringA sTrack(sKeyword.Right(sKeyword.GetLength() - 4));
          int nOffset = atoi(sTrack);
          record.m_ExtendedTrackData.SetSize(nOffset+1);
          record.m_ExtendedTrackData.ElementAt(nOffset) += MapFromProtocolString(sValue);
        }
        else if (sKeyword == "PLAYORDER")
        {
          //Handle the "PLAYORDER" line
          sPlayOrder += MapFromProtocolString(sValue);
        }
      }
    }
  }

  //Handle all the parsing errors
  if (!bFoundXMCDComment)
  {
    TRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a XMCD comment line\n"));
    return FALSE;
  }

  if (record.m_TrackOffsets.GetSize() == 0)
  {
    TRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with any Track frame offset lines\n"));
    return FALSE;
  }

  if (!bFoundDiscLength)
  {
    TRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a Disc length line\n"));
    return FALSE;
  }

  if (!bFoundSubmittedVia)
  {
    TRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a Submiteed via line\n"));
    return FALSE;
  }

  if (!bFoundDISCID)
  {
    TRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a DISCID line\n"));
    return FALSE;
  }

  if (!bFoundDTITLE)
  {
    TRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with a DTITLE line\n"));
    return FALSE;
  }

  if (record.m_TrackTitles.GetSize() != record.m_ExtendedTrackData.GetSize())
  {
    TRACE(_T("CCDDB::ParseRecordBodyHelper, CDDB server failed to return a record with an equal amount of TTITLE lines to EXTT lines\n"));
    return FALSE;
  }

  //Put together the details from the "DTITLE" line(s)
  int nFind = sTitle.Find(_T(" / "));
  if (nFind != -1)
  {
    record.m_sArtist = sTitle.Left(nFind);
    record.m_sTitle = sTitle.Right(sTitle.GetLength() - nFind - 3);
  }
  else
  {
    //Artist & Title are the same
    record.m_sTitle = sTitle;
    record.m_sArtist = sTitle;
  }

  //Handle the "PLAYORDER" line
  BOOL bContinue;
  do 
  {
    nFind = sPlayOrder.Find(_T(","));
    if (nFind != -1)
    {
      CString sNum(sPlayOrder.Left(nFind));
      int nNum = _ttoi(sNum);
      record.m_PlayOrder.Add(nNum);
      sPlayOrder = sPlayOrder.Right(sPlayOrder.GetLength() - nFind - 1);
      bContinue = TRUE;
    }
    else 
    {
      if (sPlayOrder.GetLength())
      {
        int nNum = _ttoi(sPlayOrder);
        record.m_PlayOrder.Add(nNum);
      }
      bContinue = FALSE;
    }
  }
  while (bContinue);

  return TRUE;
}

BOOL CCDDB::ParseSiteBodyHelper(CCDDBSite& site, const CStringA& sLine)
{
  //Create a local copy of sLine so that we can use strtok_s on it
  CStringA sLocalLine(sLine);    

  //Pull out "m_sSite"  
  static char* pszSeparators = " ";
  char* pszValue = sLocalLine.GetBuffer(sLocalLine.GetLength());
  char* pszNextToken = NULL;
  LPSTR pszToken = strtok_s(pszValue, pszSeparators, &pszNextToken);
  if (pszToken)
    site.m_sSite = MapFromProtocolString(pszToken);
  else
  {
    sLocalLine.ReleaseBuffer();
    return FALSE;
  }

  if (m_nProtocolVersion >= 3)
  {
    //Pull out "m_sProtocol"
    pszToken = strtok_s(NULL, pszSeparators, &pszNextToken);  
    if (pszToken)
      site.m_sProtocol = MapFromProtocolString(pszToken);
    else
    {
      sLocalLine.ReleaseBuffer();
      return FALSE;
    }
  }
  else
    site.m_sProtocol.Empty();

  //Pull out "m_nPort"
  pszToken = strtok_s(NULL, pszSeparators, &pszNextToken);  
  if (pszToken)
    site.m_nPort = atoi(pszToken);
  else
  {
    sLocalLine.ReleaseBuffer();
    return FALSE;
  }

  if (m_nProtocolVersion >= 3)
  {
    //Pull out "m_sAddress"
    pszToken = strtok_s(NULL, pszSeparators, &pszNextToken);  
    if (pszToken)
      site.m_sAddress = MapFromProtocolString(pszToken);
    else
    {
      sLocalLine.ReleaseBuffer();
      return FALSE;
    }
  }
  else
    site.m_sAddress.Empty();

  //Pull out the latitude values
  pszToken = strtok_s(NULL, pszSeparators, &pszNextToken);  
  if (pszToken)
  {
    CString sLatitude(MapFromProtocolString(pszToken));
  
    //Latitude field must be 7 characters long
    if (sLatitude.GetLength() != 7)
      return FALSE;

    //Parse out the northing field
    if (sLatitude.GetAt(0) == _T('N'))
      site.m_bNorthing = TRUE;
    else if (sLatitude.GetAt(0) == _T('S'))
      site.m_bNorthing = FALSE;
    else
    {
      sLocalLine.ReleaseBuffer();
      return FALSE;
    }

    //Parse out the Latitude degrees
    TCHAR sNum[4];
    sNum[0] = sLatitude.GetAt(1);
    sNum[1] = sLatitude.GetAt(2);
    sNum[2] = sLatitude.GetAt(3);
    sNum[3] = _T('\0');
    int nLatDeg = _ttoi(sNum);

    //Parse out the Latitude minutes
    sNum[0] = sLatitude.GetAt(5);
    sNum[1] = sLatitude.GetAt(6);
    sNum[2] = _T('\0');
    int nLatMin = _ttoi(sNum);
    site.m_nLatitudeMinutes = nLatMin + 60*nLatDeg;
  }
  else
  {
    sLocalLine.ReleaseBuffer();
    return FALSE;
  }

  //Pull out the longitude values
  pszToken = strtok_s(NULL, pszSeparators, &pszNextToken);  
  if (pszToken)
  {
    CString sLongitude(MapFromProtocolString(pszToken));
  
    //Latitude field must be 7 characters long
    if (sLongitude.GetLength() != 7)
    {
      sLocalLine.ReleaseBuffer();
      return FALSE;
    }

    //Parse out the northing field
    if (sLongitude.GetAt(0) == _T('E'))
      site.m_bEasting = TRUE;
    else if (sLongitude.GetAt(0) == _T('W'))
      site.m_bEasting = FALSE;
    else
    {
      sLocalLine.ReleaseBuffer();
      return FALSE;
    }

    //Parse out the Longitude degrees
    TCHAR sNum[4];
    sNum[0] = sLongitude.GetAt(1);
    sNum[1] = sLongitude.GetAt(2);
    sNum[2] = sLongitude.GetAt(3);
    sNum[3] = _T('\0');
    int nLongDeg = _ttoi(sNum);

    //Parse out the Longitude minutes
    sNum[0] = sLongitude.GetAt(5);
    sNum[1] = sLongitude.GetAt(6);
    sNum[2] = _T('\0');
    int nLongMin = _ttoi(sNum);
    site.m_nLongitudeMinutes = nLongMin + 60*nLongDeg;
  }
  else
  {
    sLocalLine.ReleaseBuffer();
    return FALSE;
  }

  //Whats left is put into the description field
  if (pszNextToken)
    site.m_sDescription = MapFromProtocolString(pszNextToken);
  else
    site.m_sDescription.Empty();

  sLocalLine.ReleaseBuffer();

  //Everything has been parsed correctly
  return TRUE;
}

BOOL CCDDB::ParseQueryBodyHelper(CCDDBQueryResult& result, const CStringA& sLine)
{
  //Create a local copy of sLine so that we can use strtok_s on it
  CStringA sLocalLine(sLine);    

  //Pull out the category
  static char* pszSeparators = " ";
  char* pszValue = sLocalLine.GetBuffer(sLocalLine.GetLength());
  char* pszNextToken = NULL;
  LPSTR pszToken = strtok_s(pszValue, pszSeparators, &pszNextToken);
  if (pszToken)
    result.m_sCategory = MapFromProtocolString(pszToken);
  else
  {
    sLocalLine.ReleaseBuffer();
    return FALSE;
  }

  //Pull out the CDDB DISCID
  pszToken = strtok_s(NULL, pszSeparators, &pszNextToken);  
  if (pszToken)
    sscanf_s(pszToken, "%x", &result.m_dwDiscID);
  else
  {
    sLocalLine.ReleaseBuffer();
    return FALSE;
  }

  //Pull out the Artist and Title
  if (pszNextToken)
    result.m_sTitle = MapFromProtocolString(pszNextToken);
  else
  {
    sLocalLine.ReleaseBuffer();
    return FALSE;
  }

  //Remove any EOL if it is on the end
  int nEOLPos = result.m_sTitle.Find(_T("\r\n"));
  if (nEOLPos == -1)
    nEOLPos = result.m_sTitle.Find(_T("\n"));
  if (nEOLPos != -1)
    result.m_sTitle = result.m_sTitle.Left(nEOLPos);
  
  //Split into the artist and title
  int nSlashPos = result.m_sTitle.Find(_T('/'));
  if (nSlashPos != -1)
  {
    result.m_sArtist = result.m_sTitle.Left(nSlashPos);
    result.m_sTitle = result.m_sTitle.Right(result.m_sTitle.GetLength() - nSlashPos - 1);

    //Remove any trailing or leading spaces
    result.m_sArtist.Trim();
    result.m_sTitle.Trim();
  }

  return TRUE;
}


void CCDDB::ThrowCDDBException(DWORD dwError, DWORD dwFacility, const CString& sLastResponse)
{
	if (dwError == 0)
		dwError = ::GetLastError();

	CCDDBException* pException = new CCDDBException(dwError, dwFacility, sLastResponse);

	TRACE(_T("Warning: throwing CCDDBException for error 0x%08X\n"), pException->m_hr);
	THROW(pException);
}

void CCDDB::ThrowCDDBException(HRESULT hr, const CString& sLastResponse)
{
	CCDDBException* pException = new CCDDBException(hr, sLastResponse);

	TRACE(_T("Warning: throwing CCDDBException for error 0x%08X\n"), pException->m_hr);
	THROW(pException);
}

void CCDDB::ThrowCDDBMMException(DWORD dwError, const CString& sLastResponse)
{
	CCDDBException* pException = new CCDDBException(TRUE, dwError, sLastResponse);

	TRACE(_T("Warning: throwing CCDDBException for error 0x%08X\n"), pException->m_hr);
	THROW(pException);
}
